#!/usr/bin/env python3
"""
Visualization Generation Module for Dane County School Dashboard

This module creates heatmap visualizations from processed BioMarkX report card data.
Creates separate heatmaps for all schools, elementary schools, middle schools, and high schools.
"""

import sys
import os
import tomllib
from pathlib import Path
from typing import Dict, List, Any, Optional, Tuple
import pandas as pd
import numpy as np
import plotly.graph_objects as go
from datetime import datetime, timedelta
from rich.console import Console
from rich.progress import Progress, SpinnerColumn, TextColumn, BarColumn
import argparse

# Load environment variables from .env file if running locally
try:
    from dotenv import load_dotenv
    # Look for .env file in the project root
    env_path = Path(__file__).parent.parent.parent / '.env'
    if env_path.exists():
        load_dotenv(env_path)
except ImportError:
    # dotenv not available, rely on environment variables being set externally
    pass

# Import our data modules
from create_report_card_data import create_report_card_data, get_report_card_summary
from retrieve_virus_data import retrieve_school_virus_data

# Environment Detection (duplicated from create_discussions.py - these functions are not used here)

# Visual constants for school heatmaps
SCHOOL_TYPE_COLORS = {
    'elementary': '#E74C3C',    # Red
    'middle': '#3498DB',        # Blue  
    'high': '#2ECC71',          # Green
    'all': '#9B59B6'            # Purple
}

# Heatmap colorscale (white for 0, dark blue for 1) - improved visibility for 1-25% range
COLORSCALE = [
    [0.0, 'white'],           # 0% positive (no detections)
    [0.01, '#f5fbff'],        # 1% - very faint blue
    [0.05, '#e6f3ff'],        # 5% - light blue
    [0.10, '#cce6ff'],        # 10% - light blue
    [0.15, '#b3daff'],        # 15% - light-medium blue
    [0.20, '#99ccff'],        # 20% - medium light blue
    [0.25, '#80bfff'],        # 25% - medium blue
    [0.3, '#66b2ff'],         # 30% - medium blue
    [0.4, '#4da6ff'],         # 40% - medium blue
    [0.5, '#3399ff'],         # 50% - medium blue
    [0.6, '#1a8cff'],         # 60% - medium-dark blue
    [0.7, '#0066cc'],         # 70% - dark blue
    [0.8, '#004d99'],         # 80% - darker blue
    [0.9, '#003366'],         # 90% - very dark blue
    [1.0, '#001a33']          # 100% - darkest blue
]

def load_sites_config(sites_path: Optional[Path] = None) -> Dict[str, Any]:
    """Load sites configuration from sites.toml file."""
    if sites_path is None:
        sites_path = Path(__file__).parent / "sites.toml"
    else:
        sites_path = Path(sites_path)
    
    if not sites_path.exists():
        raise FileNotFoundError(f"Sites configuration file not found: {sites_path}")
    
    try:
        with open(sites_path, 'rb') as f:
            config = tomllib.load(f)
        return config
    except Exception as e:
        raise Exception(f"Failed to load sites configuration from {sites_path}: {e}")

def get_ordered_virus_names(virus_list: List[str]) -> List[str]:
    """
    Order virus names according to specified priority:
    1. SARS-CoV-2 (top)
    2. pan-Influenza A
    3. Influenza A H1N1
    4. Influenza A H3N2
    5. Influenza B
    6. pan-RSV
    7. RSV-A
    8. RSV-B
    9. Human coronaviruses (229E, HKU1, NL63, OC43)
    10. Other viruses (alphabetical)
    11. no-assay-control (negative control)
    12. RNAseP (positive control, bottom)
    """
    # Define priority order
    priority_order = [
        'SARS-CoV-2',
        'pan-Influenza A', 
        'Influenza A H1N1',
        'Influenza A H3N2',
        'Influenza B',
        'pan-RSV',
        'RSV-A',
        'RSV-B',
        'Coronavirus 229E',
        'Coronavirus HKU1', 
        'Coronavirus NL63',
        'Coronavirus OC43'
    ]
    
    # Controls at the end
    controls = ['no-assay-control', 'RNAseP']
    
    # Control display names for visualization
    control_display_names = {
        'no-assay-control': 'No primers (negative control)',
        'RNAseP': 'RNAseP (positive control)'
    }
    
    # Separate viruses into categories
    priority_viruses = []
    other_viruses = []
    control_viruses = []
    
    for virus in virus_list:
        if virus in priority_order:
            priority_viruses.append(virus)
        elif virus in controls:
            control_viruses.append(virus)
        else:
            other_viruses.append(virus)
    
    # Sort each category
    priority_viruses.sort(key=lambda x: priority_order.index(x))
    other_viruses.sort()  # Alphabetical
    
    # Put controls in specific order (negative control first, then positive control)
    ordered_controls = []
    if 'no-assay-control' in control_viruses:
        ordered_controls.append('No primers (negative control)')
    if 'RNAseP' in control_viruses:
        ordered_controls.append('RNAseP (positive control)')
    
    # Apply display names to other controls if any
    display_controls = []
    for control in control_viruses:
        if control in control_display_names:
            display_controls.append(control_display_names[control])
        else:
            display_controls.append(control)
    
    # Use the ordered controls we built above
    final_controls = ordered_controls
    
    # Combine all categories
    return priority_viruses + other_viruses + final_controls

def create_monday_friday_weeks(df: pd.DataFrame) -> pd.DataFrame:
    """
    Create Monday-Friday week ranges based on start_time.
    """
    df = df.copy()
    
    # Ensure start_time is datetime
    if not pd.api.types.is_datetime64_any_dtype(df['start_time']):
        df['start_time'] = pd.to_datetime(df['start_time'], errors='coerce')
    
    # Find the Monday of each week (start of school week)
    df['week_monday'] = df['start_time'].dt.to_period('W-MON').dt.start_time
    
    # Calculate the Friday of each week (end of school week)
    df['week_friday'] = df['week_monday'] + pd.Timedelta(days=4)
    
    # Create week label in Monday-Friday format
    df['school_week_range'] = df.apply(
        lambda row: f"{row['week_monday'].strftime('%Y-%m-%d')} to {row['week_friday'].strftime('%Y-%m-%d')}", 
        axis=1
    )
    
    return df

def create_heatmap_data(df: pd.DataFrame, school_type_filter: str = 'all', academic_year_filter: str = None) -> Tuple[np.ndarray, List[List[str]], List[str], List[str]]:
    """
    Create heatmap data matrix from processed report card data.
    
    Args:
        df: Processed report card DataFrame
        school_type_filter: 'all', 'elementary', 'middle', or 'high'
        academic_year_filter: Specific academic year (e.g., '2024-25') or None for all years
        
    Returns:
        Tuple of (z_values, hover_text, virus_names, week_labels)
    """
    # Filter by school type if specified
    if school_type_filter != 'all':
        if 'school_type' not in df.columns:
            raise ValueError("school_type column not found in DataFrame")
        df = df[df['school_type'] == school_type_filter].copy()
    
    # Filter by academic year if specified
    if academic_year_filter:
        if 'academic_year' not in df.columns:
            raise ValueError("academic_year column not found in DataFrame")
        df = df[df['academic_year'] == academic_year_filter].copy()
    
    if df.empty:
        return np.array([]), [[]], [], []
    
    # Create Monday-Friday school weeks
    df = create_monday_friday_weeks(df)
    
    # Filter to valid test results only
    valid_results = df[df['qualitative_result'].isin(['Positive', 'Negative'])].copy()
    
    if valid_results.empty:
        return np.array([]), [[]], [], []
    
    # Get unique viruses
    all_virus_names = valid_results['normalized_assay_target'].unique()
    
    # Generate all weeks in the date range (including weeks with no data)
    if not valid_results.empty:
        # Find the earliest and latest dates in the data
        earliest_date = valid_results['week_monday'].min()
        latest_date = valid_results['week_monday'].max()
        
        # Generate all Monday-Friday week ranges
        all_weeks = []
        current_monday = earliest_date
        
        while current_monday <= latest_date:
            current_friday = current_monday + pd.Timedelta(days=4)
            week_label = f"{current_monday.strftime('%Y-%m-%d')} to {current_friday.strftime('%Y-%m-%d')}"
            all_weeks.append(week_label)
            current_monday += pd.Timedelta(days=7)
        
        week_labels = all_weeks
    else:
        week_labels = []
    
    # Create mapping for display names
    control_display_names = {
        'no-assay-control': 'No primers (negative control)',
        'RNAseP': 'RNAseP (positive control)'
    }
    
    # Order viruses according to priority (includes controls)
    virus_names = get_ordered_virus_names(all_virus_names)
    
    if not virus_names or not week_labels:
        return np.array([]), [[]], [], []
    
    # Initialize data matrices
    n_viruses = len(virus_names)
    n_weeks = len(week_labels)
    z_values = np.full((n_viruses, n_weeks), np.nan)
    hover_text = [['' for _ in range(n_weeks)] for _ in range(n_viruses)]
    
    # Create reverse mapping to find original names from display names
    display_to_original = {v: k for k, v in control_display_names.items()}
    
    # Calculate frequencies for each virus-week combination
    for i, virus_display in enumerate(virus_names):
        for j, week in enumerate(week_labels):
            # Get the original virus name for data lookup
            virus_original = display_to_original.get(virus_display, virus_display)
            
            # Filter to this virus and week
            virus_week_data = valid_results[
                (valid_results['normalized_assay_target'] == virus_original) & 
                (valid_results['school_week_range'] == week)
            ]
            
            if not virus_week_data.empty:
                # Calculate statistics
                total_samples = len(virus_week_data)
                positive_samples = len(virus_week_data[virus_week_data['qualitative_result'] == 'Positive'])
                negative_samples = total_samples - positive_samples
                unique_schools = virus_week_data['facility_name'].nunique()
                
                # Calculate positive frequency (0.0 to 1.0)
                positive_frequency = positive_samples / total_samples if total_samples > 0 else 0.0
                z_values[i, j] = positive_frequency
                
                # Create hover text with date range (use display name)
                hover_text[i][j] = (
                    f"<b>{virus_display}</b><br>"
                    f"Date range: {week}<br>"
                    f"Schools: {unique_schools}<br>"
                    f"Total samples: {total_samples}<br>"
                    f"Positive: {positive_samples}<br>"
                    f"Negative: {negative_samples}<br>"
                    f"Positive rate: {positive_frequency:.1%}"
                )
            else:
                # No data for this virus-week combination - leave NaN to show as missing
                # z_values[i, j] is already NaN from initialization
                hover_text[i][j] = f"<b>{virus_display}</b><br>Date range: {week}<br>No data collected"
    
    return z_values, hover_text, virus_names, week_labels

def create_school_heatmap(df: pd.DataFrame, school_type: str, academic_year: str, output_dir: Path, 
                         title_prefix: str = "", generate_html: bool = True) -> bool:
    """
    Create a heatmap visualization for school virus surveillance data.
    
    Args:
        df: Processed report card DataFrame
        school_type: 'all', 'elementary', 'middle', or 'high'
        academic_year: Academic year (e.g., '2024-25')
        output_dir: Directory to save outputs
        title_prefix: Prefix for file names
        generate_html: Whether to generate HTML files
        
    Returns:
        True if successful, False otherwise
    """
    console = Console()
    
    try:
        # Create heatmap data
        z_values, hover_text, virus_names, week_labels = create_heatmap_data(df, school_type, academic_year)
        
        if len(virus_names) == 0 or len(week_labels) == 0:
            console.print(f"[yellow]No data available for {school_type} schools in {academic_year}[/yellow]")
            return False
        
        # Create file suffix (no title needed since we're removing titles from plots)
        if school_type == 'all':
            file_suffix = f"all_schools_{academic_year.replace('-', '_')}"
        else:
            file_suffix = f"{school_type}_schools_{academic_year.replace('-', '_')}"
        
        # Create the heatmap
        fig = go.Figure(data=go.Heatmap(
            z=z_values,
            x=week_labels,
            y=virus_names,
            hovertemplate='%{customdata}<extra></extra>',
            customdata=hover_text,
            colorscale=COLORSCALE,
            zmin=0,
            zmax=1,
            colorbar=dict(
                title=dict(text="Positive Rate", side="right"),
                tickmode="array",
                tickvals=[0, 0.25, 0.5, 0.75, 1.0],
                ticktext=["0%", "25%", "50%", "75%", "100%"]
            ),
            hoverongaps=False
        ))
        
        # Create selective X-axis labels (show every 4th week to reduce clutter)
        x_tick_indices = list(range(0, len(week_labels), 4))  # Every 4th week
        if len(week_labels) - 1 not in x_tick_indices:  # Always include last week
            x_tick_indices.append(len(week_labels) - 1)
        
        x_tick_labels = []
        x_tick_values = []
        for i in x_tick_indices:
            if i < len(week_labels):
                # Extract just the start date (Monday) for cleaner display
                week_range = week_labels[i]
                start_date = week_range.split(' to ')[0]
                x_tick_labels.append(start_date)
                x_tick_values.append(i)
        
        # Update layout
        fig.update_layout(
            xaxis=dict(
                title="Week",
                tickmode='array',
                tickvals=x_tick_values,
                ticktext=x_tick_labels,
                tickangle=45,
                tickfont=dict(size=10),
                title_font=dict(size=14)
            ),
            yaxis=dict(
                title="Virus",
                tickfont=dict(size=12),
                title_font=dict(size=14)
            ),
            width=max(1200, len(week_labels) * 30),  # Dynamic width based on weeks (50% wider)
            height=max(600, len(virus_names) * 30),  # Dynamic height based on viruses
            margin=dict(l=200, r=150, t=80, b=120),  # More left margin for virus names
            font=dict(family="Arial, sans-serif"),
            plot_bgcolor='white'
        )
        
        # Save outputs
        output_dir.mkdir(parents=True, exist_ok=True)
        
        # Create file paths
        if title_prefix:
            base_name = f"{title_prefix}_{file_suffix}"
        else:
            base_name = file_suffix
        
        # Save PNG
        png_path = output_dir / f"{base_name}_heatmap.png"
        fig.write_image(str(png_path), width=fig.layout.width, height=fig.layout.height, scale=2)
        console.print(f"[green]✓ Saved PNG: {png_path.name}[/green]")
        
        # Save HTML if requested
        if generate_html:
            html_path = output_dir / f"{base_name}_heatmap.html"
            fig.write_html(str(html_path))
            console.print(f"[green]✓ Saved HTML: {html_path.name}[/green]")
        
        return True
        
    except Exception as e:
        console.print(f"[red]✗ Failed to create {school_type} schools heatmap: {e}[/red]")
        return False

def create_all_heatmaps(df: pd.DataFrame, output_dir: Path, title_prefix: str = "", 
                       generate_html: bool = True) -> Dict[str, bool]:
    """
    Create all heatmaps: all schools, elementary, middle, and high schools.
    Subdivided by academic year.
    
    Args:
        df: Processed report card DataFrame
        output_dir: Directory to save outputs
        title_prefix: Prefix for file names
        generate_html: Whether to generate HTML files
        
    Returns:
        Dictionary with success status for each heatmap
    """
    console = Console()
    results = {}
    
    # Get unique academic years - filter to 2024-25 only
    if 'academic_year' not in df.columns:
        console.print("[red]❌ academic_year column not found in DataFrame[/red]")
        return {}
    
    # Filter to 2024-25 academic year only
    academic_years = ['2024-25']
    
    # Define school types to process
    school_types = [
        ('all', 'All Schools'),
        ('elementary', 'Elementary Schools'),
        ('middle', 'Middle Schools'), 
        ('high', 'High Schools')
    ]
    
    total_combinations = len(academic_years) * len(school_types)
    console.print(f"[blue]Creating heatmaps for {len(academic_years)} academic years × {len(school_types)} school categories = {total_combinations} total heatmaps[/blue]")
    console.print(f"[cyan]Academic years: {academic_years}[/cyan]")
    
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        BarColumn(),
        TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
        console=console
    ) as progress:
        
        task = progress.add_task("Creating heatmaps...", total=total_combinations)
        
        for academic_year in academic_years:
            for school_type, description in school_types:
                progress.update(task, description=f"Creating {description} heatmap for {academic_year}...")
                
                success = create_school_heatmap(
                    df=df,
                    school_type=school_type,
                    academic_year=academic_year,
                    output_dir=output_dir,
                    title_prefix=title_prefix,
                    generate_html=generate_html
                )
                
                results[f"{school_type}_{academic_year}"] = success
                progress.advance(task)
    
    # Summary
    successful = sum(results.values())
    console.print(f"[green]✓ Successfully created {successful}/{total_combinations} heatmaps[/green]")
    
    # Show breakdown by academic year
    for academic_year in academic_years:
        year_results = [results[f"{school_type}_{academic_year}"] for school_type, _ in school_types]
        year_successful = sum(year_results)
        console.print(f"  [cyan]{academic_year}: {year_successful}/{len(school_types)} heatmaps[/cyan]")
    
    return results

def main():
    """Main function for creating school virus surveillance heatmaps."""
    parser = argparse.ArgumentParser(description="Generate school virus surveillance heatmaps")
    parser.add_argument("--input-file", type=str, help="Path to processed report card Excel file")
    parser.add_argument("--output-dir", type=str, help="Output directory for visualizations")
    parser.add_argument("--title-prefix", type=str, default="", help="Prefix for output file names")
    parser.add_argument("--no-html", action="store_true", help="Skip HTML generation (PNG only)")
    parser.add_argument("--test-mode", action="store_true", help="Use test data subset")
    
    args = parser.parse_args()
    
    console = Console()
    console.print("[bold blue]🏫 Dane County School Virus Surveillance Heatmap Generator[/bold blue]\n")
    
    try:
        # Determine output directory
        if args.output_dir:
            output_dir = Path(args.output_dir)
        else:
            # Use local outputs by default
            script_dir = Path(__file__).parent
            if os.getenv('GITHUB_ACTIONS'):
                base_dir = script_dir / "github-actions-outputs"
                output_dir = base_dir / "visualizations"
            else:
                base_dir = script_dir / "local-outputs"
                output_dir = base_dir / "heatmaps"
        
        # Load data
        if args.input_file:
            # Load from specified Excel file
            console.print(f"[cyan]Loading data from: {args.input_file}[/cyan]")
            df = pd.read_excel(args.input_file, sheet_name='Report Card Data')
        else:
            # Generate fresh data
            console.print("[cyan]Generating fresh data from LabKey...[/cyan]")
            if args.test_mode:
                raw_df = retrieve_school_virus_data(max_rows=1000, show_progress=True, test_filter=True)
            else:
                raw_df = retrieve_school_virus_data(show_progress=True)
            
            if raw_df.empty:
                console.print("[red]❌ No data retrieved from LabKey[/red]")
                return 1
            
            df = create_report_card_data(raw_df, show_progress=True)
        
        if df.empty:
            console.print("[red]❌ No processed data available[/red]")
            return 1
        
        # Display data summary
        summary = get_report_card_summary(df)
        console.print(f"\n[cyan]📊 Data Summary:[/cyan]")
        console.print(f"  Records: {summary['total_records']:,}")
        console.print(f"  Schools: {summary['unique_schools']}")
        console.print(f"  Viruses: {summary['unique_targets']}")
        console.print(f"  Academic years: {summary['academic_years']}")
        
        # Show school type breakdown
        if 'school_type' in df.columns:
            school_type_counts = df['school_type'].value_counts()
            console.print(f"  School types: {dict(school_type_counts)}")
        
        # Create heatmaps
        console.print(f"\n[cyan]Creating heatmaps...[/cyan]")
        
        results = create_all_heatmaps(
            df=df,
            output_dir=output_dir,
            title_prefix=args.title_prefix,
            generate_html=not args.no_html
        )
        
        # Display results
        console.print(f"\n[green]✅ Heatmap generation complete![/green]")
        console.print(f"[cyan]Output directory: {output_dir}[/cyan]")
        
        for school_type, success in results.items():
            status = "✓" if success else "✗"
            color = "green" if success else "red"
            console.print(f"  [{color}]{status} {school_type.title()} schools heatmap[/{color}]")
        
        return 0
        
    except Exception as e:
        console.print(f"[red]❌ Error: {e}[/red]")
        import traceback
        traceback.print_exc()
        return 1

if __name__ == "__main__":
    exit_code = main()
    sys.exit(exit_code)